@@ -106,17 +106,16 @@ fn format_cache_control(cc: &CacheControl) -> String {
     out
 }
 
-/// Holds configuration options which control the behavior of the cache and
-/// are independent of any specific request or response.
+/// Configuration options which control behavior of the cache. Use with `CachePolicy::new_options()`.
 #[derive(Debug, Copy, Clone)]
 #[cfg_attr(feature = "with_serde", derive(serde_derive::Serialize, serde_derive::Deserialize))]
-pub struct CachePolicyOptions {
-    /// If `shared` is `true` (default), then the response is evaluated from a
+pub struct CacheOptions {
+    /// If `true` (default), then the response is evaluated from a
     /// perspective of a shared cache (i.e. `private` is not cacheable and
     /// `s-maxage` is respected). If `shared` is `false`, then the response is
     /// evaluated from a perspective of a single-user cache (i.e. `private` is
-    /// cacheable and `s-maxage` is ignored). `shared: true` is recommended
-    /// for HTTP clients.
+    /// cacheable and `s-maxage` is ignored). `shared: true` is required
+    /// for proxies and multi-user caches.
     pub shared: bool,
     /// `cache_heuristic` is a fraction of response's age that is used as a
     /// fallback cache duration. The default is 0.1 (10%), e.g. if a file
@@ -134,24 +133,15 @@ pub struct CachePolicyOptions {
     /// found in bad StackOverflow answers and PHP's "session limiter"
     /// defaults.
     pub ignore_cargo_cult: bool,
-    /// If `trust_server_date` is `false`, then server's `Date` header won't
-    /// be used as the base for `max-age`. This is against the RFC, but it's
-    /// useful if you want to cache responses with very short `max-age`, but
-    /// your local clock is not exactly in sync with the server's.
-    pub trust_server_date: bool,
-    /// When the response has been received. `SystemTime::now`.
-    pub response_time: SystemTime,
 }
 
-impl Default for CachePolicyOptions {
+impl Default for CacheOptions {
     fn default() -> Self {
         Self {
             shared: true,
             cache_heuristic: 0.1, // 10% matches IE
             immutable_min_time_to_live: Duration::from_secs(24 * 3600),
             ignore_cargo_cult: false,
-            trust_server_date: true,
-            response_time: SystemTime::now(),
         }
     }
 }
@@ -173,26 +163,49 @@ pub struct CachePolicy {
     status: StatusCode,
     #[cfg_attr(feature = "with_serde", serde(with = "http_serde::method"))]
     method: Method,
-    opts: CachePolicyOptions,
+    opts: CacheOptions,
     res_cc: CacheControl,
     req_cc: CacheControl,
+    response_time: SystemTime,
 }
 
 impl CachePolicy {
     /// Cacheability of an HTTP response depends on how it was requested, so
     /// both request and response are required to create the policy.
     #[inline]
-    pub fn new<Req: RequestLike, Res: ResponseLike>(
+    pub fn new<Req: RequestLike, Res: ResponseLike>(req: &Req, res: &Res) -> Self {
+        let uri = req.uri();
+        let status = res.status();
+        let method = req.method().clone();
+        let res = res.headers().clone();
+        let req = req.headers().clone();
+        Self::from_details(
+            uri,
+            method,
+            status,
+            req,
+            res,
+            SystemTime::now(),
+            Default::default(),
+        )
+    }
+
+    /// Caching with customized behavior. See `CacheOptions` for details.
+    ///
+    /// `response_time` is a timestamp when the response has been received, usually `SystemTime::now()`.
+    #[inline]
+    pub fn new_options<Req: RequestLike, Res: ResponseLike>(
         req: &Req,
         res: &Res,
-        opts: CachePolicyOptions,
+        response_time: SystemTime,
+        opts: CacheOptions,
     ) -> Self {
         let uri = req.uri();
         let status = res.status();
         let method = req.method().clone();
         let res = res.headers().clone();
         let req = req.headers().clone();
-        Self::from_details(uri, method, status, req, res, opts)
+        Self::from_details(uri, method, status, req, res, response_time, opts)
     }
 
     fn from_details(
@@ -201,7 +214,8 @@ impl CachePolicy {
         status: StatusCode,
         req: HeaderMap,
         mut res: HeaderMap,
-        opts: CachePolicyOptions,
+        response_time: SystemTime,
+        opts: CacheOptions,
     ) -> Self {
         let mut res_cc = parse_cache_control(res.get_all("cache-control"));
         let req_cc = parse_cache_control(req.get_all("cache-control"));
@@ -244,6 +258,7 @@ impl CachePolicy {
             req_cc,
             opts,
             uri,
+            response_time,
         }
     }
 
@@ -298,12 +313,28 @@ impl CachePolicy {
     /// (e.g. it's for a different URL or method), or may require to be
     /// refreshed first. Either way, the new request's headers will have been
     /// updated for sending it to the origin server.
-    pub fn satisfies_without_revalidation<Req: RequestLike>(
-        &self,
-        req: &Req,
-        now: SystemTime,
-    ) -> bool {
+    pub fn before_request<Req: RequestLike>(&self, req: &Req, now: SystemTime) -> BeforeRequest {
         let req_headers = req.headers();
+
+        // revalidation allowed via HEAD
+        let (exact_match, may_revalidate) = self.request_matches(req);
+
+        if exact_match && self.satisfies_without_revalidation(&req_headers, now) {
+            BeforeRequest::Fresh(self.cached_response(now))
+        } else if may_revalidate {
+            BeforeRequest::Stale {
+                request: self.revalidation_request(req),
+                matches: false,
+            }
+        } else {
+            BeforeRequest::Stale {
+                request: self.request_from_headers(req_headers.clone()),
+                matches: false,
+            }
+        }
+    }
+
+    fn satisfies_without_revalidation(&self, req_headers: &HeaderMap, now: SystemTime) -> bool {
         // When presented with a request, a cache MUST NOT reuse a stored response, unless:
         // the presented request does not contain the no-cache pragma (Section 5.4), nor the no-cache cache directive,
         // unless the stored response is successfully validated (Section 4.3), and
@@ -355,19 +386,20 @@ impl CachePolicy {
             }
         }
 
-        self.request_matches(req, false)
+        true
     }
 
-    fn request_matches<Req: RequestLike>(&self, req: &Req, allow_head_method: bool) -> bool {
+    /// returns: matches including method, matches allowing head
+    fn request_matches<Req: RequestLike>(&self, req: &Req) -> (bool, bool) {
         // The presented effective request URI and that of the stored response match, and
-        req.is_same_uri(&self.uri) &&
-            self.req.get("host") == req.headers().get("host") &&
-            // the request method associated with the stored response allows it to be used for the presented request, and
-            (
-                self.method == req.method() ||
-                (allow_head_method && Method::HEAD == req.method())) &&
+        let matches = req.is_same_uri(&self.uri) &&
+            (self.req.get("host") == req.headers().get("host")) &&
             // selecting header fields nominated by the stored response (if any) match those presented, and
-            self.vary_matches(req)
+            self.vary_matches(req);
+        let exact_match = matches && self.method == req.method();
+
+        // the request method associated with the stored response allows it to be used for the presented request, and
+        (exact_match, exact_match || Method::HEAD == req.method())
     }
 
     fn allows_storing_authenticated(&self) -> bool {
@@ -426,7 +458,7 @@ impl CachePolicy {
     ///
     /// It returns response "parts" without a body. You can upgrade it to a full
     /// response with `Response::from_parts(parts, BYOB)`
-    pub fn cached_response(&self, now: SystemTime) -> http::response::Parts {
+    fn cached_response(&self, now: SystemTime) -> http::response::Parts {
         let mut headers = Self::copy_without_hop_by_hop_headers(&self.res);
         let age = self.age(now);
         let day = Duration::from_secs(3600 * 24);
@@ -459,17 +491,7 @@ impl CachePolicy {
         parts
     }
 
-    /// Value of the Date response header or current time if Date was demed invalid
-    ///
-    fn date(&self) -> SystemTime {
-        if self.opts.trust_server_date {
-            self.server_date()
-        } else {
-            self.opts.response_time
-        }
-    }
-
-    fn server_date(&self) -> SystemTime {
+    fn raw_server_date(&self) -> SystemTime {
         let date = self
             .res
             .get_str("date")
@@ -477,35 +499,22 @@ impl CachePolicy {
             .and_then(|d| {
                 SystemTime::UNIX_EPOCH.checked_add(Duration::from_secs(d.timestamp() as _))
             });
-        if let Some(date) = date {
-            let max_clock_drift = Duration::from_secs(8 * 3600);
-            let clock_drift = if self.opts.response_time > date {
-                self.opts.response_time.duration_since(date)
-            } else {
-                date.duration_since(self.opts.response_time)
-            }
-            .unwrap();
-            if clock_drift < max_clock_drift {
-                return date;
-            }
-        }
-        self.opts.response_time
+        return date.unwrap_or(self.response_time)
     }
 
-    /// Value of the Age header, in seconds, updated for the current time.
+    /// Tells how long the response has been sitting in cache(s).
+    ///
+    /// Value of the `Age` header, updated for the current time.
     pub fn age(&self, now: SystemTime) -> Duration {
-        let mut age = self.age_header();
-        if let Ok(since_date) = self.opts.response_time.duration_since(self.date()) {
-            age = age.max(since_date);
-        }
+        let mut age = self.age_header_value();
 
-        if let Ok(resident_time) = now.duration_since(self.opts.response_time) {
+        if let Ok(resident_time) = now.duration_since(self.response_time) {
             age += resident_time;
         }
         age
     }
 
-    fn age_header(&self) -> Duration {
+    fn age_header_value(&self) -> Duration {
         Duration::from_secs(
             self.res
                 .get_str("age")
@@ -514,10 +523,12 @@ impl CachePolicy {
         )
     }
 
-    /// Value of applicable max-age (or heuristic equivalent) in seconds. This counts since response's `Date`.
+    /// Value of applicable max-age (or heuristic equivalent) in seconds.
+    ///
+    /// This counts since response's `Date` - `Age`.
     ///
     /// For an up-to-date value, see `time_to_live()`.
-    pub fn max_age(&self) -> Duration {
+    fn max_age(&self) -> Duration {
         if !self.is_storable() || self.res_cc.contains_key("no-cache") {
             return Duration::from_secs(0);
         }
@@ -557,7 +568,7 @@ impl CachePolicy {
             Duration::from_secs(0)
         };
 
-        let server_date = self.server_date();
+        let server_date = self.raw_server_date();
         if let Some(expires) = self.res.get_str("expires") {
             return match DateTime::parse_from_rfc2822(expires) {
                 // A cache recipient MUST interpret invalid date formats, especially the value "0", as representing a time in the past (i.e., "already expired").
@@ -591,7 +602,7 @@ impl CachePolicy {
     /// After that time (when `time_to_live() <= 0`) the response might not be
     /// usable without revalidation. However, there are exceptions, e.g. a
     /// client can explicitly allow stale responses, so always check with
-    /// `satisfies_without_revalidation()`.
+    /// `before_request()`.
     pub fn time_to_live(&self, now: SystemTime) -> Duration {
         self.max_age()
             .checked_sub(self.age(now))
@@ -611,14 +622,13 @@ impl CachePolicy {
     ///
     /// It returns request "parts" without a body. You can upgrade it to a full
     /// response with `Request::from_parts(parts, BYOB)` (the body is usually `()`).
-    pub fn revalidation_request<Req: RequestLike>(&self, incoming_req: &Req) -> http::request::Parts {
+    fn revalidation_request<Req: RequestLike>(&self, incoming_req: &Req) -> http::request::Parts {
         let mut headers = Self::copy_without_hop_by_hop_headers(incoming_req.headers());
 
         // This implementation does not understand range requests
         headers.remove("if-range");
 
-        if !self.request_matches(incoming_req, true) || !self.is_storable() {
-            // revalidation allowed via HEAD
+        if !self.is_storable() {
             // not for the same resource, or wasn't allowed to be cached anyway
             headers.remove("if-none-match");
             headers.remove("if-modified-since");
@@ -678,12 +688,12 @@ impl CachePolicy {
     ///
     /// Returns `{policy, modified}` where modified is a boolean indicating
     /// whether the response body has been modified, and old cached body can't be used.
-    pub fn revalidated_policy<Req: RequestLike, Res: ResponseLike>(
+    pub fn after_response<Req: RequestLike, Res: ResponseLike>(
         &self,
         request: &Req,
         response: &Res,
-        response_time: SystemTime
-    ) -> RevalidatedPolicy {
+        response_time: SystemTime,
+    ) -> AfterResponse {
         let response_headers = response.headers();
         let mut response_status = response.status();
 
@@ -723,7 +733,6 @@ impl CachePolicy {
             }
         }
 
-        let modified = response.status() != StatusCode::NOT_MODIFIED;
         let new_response_headers = if matches {
             let mut new_response_headers = HeaderMap::with_capacity(self.res.keys_len());
             // use other header fields provided in the 304 (Not Modified) response to replace all instances
@@ -744,32 +753,31 @@ impl CachePolicy {
             response_headers.clone()
         };
 
-        let mut new_response = Response::builder().status(response_status).body(()).unwrap().into_parts().0;
-        new_response.headers = new_response_headers;
-        let mut new_options = self.opts;
-        new_options.response_time = response_time;
+        let new_policy = CachePolicy::from_details(
+            request.uri(),
+            request.method().clone(),
+            response_status,
+            request.headers().clone(),
+            new_response_headers,
+            response_time,
+            self.opts,
+        );
+        let new_response = new_policy.cached_response(response_time);
 
-        RevalidatedPolicy {
-            policy: CachePolicy::new(request, &new_response, new_options),
-            // Client receiving 304 without body, even if it's invalid/mismatched has no option
-            // but to reuse a cached body. We don't have a good way to tell clients to do
-            // error recovery in such case.
-            modified,
-            matches,
+        if matches && response.status() == StatusCode::NOT_MODIFIED {
+            AfterResponse::NotModified(new_policy, new_response)
+        } else {
+            AfterResponse::Modified(new_policy, new_response)
         }
     }
 }
 
-/// New policy and flags
-pub struct RevalidatedPolicy {
-    /// New policy with new headers
-    pub policy: CachePolicy,
-    /// Boolean indicating whether the response body has changed.
-    /// If `false`, then a valid 304 Not Modified response has been received, and you can reuse the old cached response body.
-    /// If `true`, you should use new response's body (if present), or make another request to the origin server without any conditional headers (i.e. don't use `revalidation_headers()` this time) to get the new resource.
-    pub modified: bool,
-    /// This was a request for something else and can't be used?
-    pub matches: bool,
+/// New policy and flags to act on `after_response()`
+pub enum AfterResponse {
+    /// You can use the cached body! Make sure to use these updated headers
+    NotModified(CachePolicy, http::response::Parts),
+    /// You need to update the body in the cache
+    Modified(CachePolicy, http::response::Parts),
 }
 
 fn get_all_comma<'a>(
@@ -803,7 +811,32 @@ fn join<'a>(parts: impl Iterator<Item = &'a str>) -> String {
     out
 }
 
-/// Allows using either Request or request::Parts, or your own newtype.
+/// Next action suggested after `before_request()`
+pub enum BeforeRequest {
+    /// Good news! You can use it with body from the cache. No need to contact the server.
+    Fresh(http::response::Parts),
+    /// You must send the request to the server first.
+    Stale {
+        /// Send this request to the server (it has added revalidation headers when appropriate)
+        request: http::request::Parts,
+        /// If `false`, request was for some other resource that isn't
+        /// semantically the same as previously cached request+response
+        matches: bool,
+    },
+}
+
+impl BeforeRequest {
+    /// For backwards compatibility only.
+    /// Don't forget to use request headers from `BeforeRequest::Fresh`
+    pub fn satisfies_without_revalidation(&self) -> bool {
+        match self {
+            Self::Fresh(_) => true,
+            _ => false,
+        }
+    }
+}
+
+/// Allows using either `Request` or `request::Parts`, or your own newtype.
 pub trait RequestLike {
     /// Same as `req.uri().clone()`
     fn uri(&self) -> Uri;
@@ -817,7 +850,7 @@ pub trait RequestLike {
     fn headers(&self) -> &HeaderMap;
 }
 
-/// Allows using either Response or response::Parts, or your own newtype.
+/// Allows using either `Response` or `response::Parts`, or your own newtype.
 pub trait ResponseLike {
     /// Same as `res.status()`
     fn status(&self) -> StatusCode;
